@@ -38,17 +38,7 @@ class Missing:
 MISSING = Missing()
 
 T = TypeVar("T")
-if sys.version_info >= (3, 5, 3):
-    MaybeMissing = Union[T, Missing]
-else:
-    # work around a bug in 3.5.2 and earlier's typing.py
-    class MaybeMissingMeta(type):
-        def __getitem__(self, arg: Any) -> Any:
-            return Union[arg, Missing]
-
-    class MaybeMissing(metaclass=MaybeMissingMeta):  # type: ignore
-        pass
-
+MaybeMissing = Union[T, Missing]
 
 _formatter = FancyFormatter(sys.stdout, sys.stderr, False)
 
@@ -107,7 +97,7 @@ class Error:
             return _style(self.object_desc, bold=True) + " " + self.message
 
         stub_line = None
-        stub_file = None  # type: None
+        stub_file: None = None
         if not isinstance(self.stub_object, Missing):
             stub_line = self.stub_object.line
         # TODO: Find a way of getting the stub file
@@ -171,6 +161,10 @@ def test_module(module_name: str) -> Iterator[Error]:
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
             runtime = importlib.import_module(module_name)
+            # Also run the equivalent of `from module import *`
+            # This could have the additional effect of loading not-yet-loaded submodules
+            # mentioned in __all__
+            __import__(module_name, fromlist=["*"])
     except Exception as e:
         yield Error([module_name], "failed to import: {}".format(e), stub, MISSING)
         return
@@ -182,7 +176,7 @@ def test_module(module_name: str) -> Iterator[Error]:
 
 @singledispatch
 def verify(
-    stub: nodes.Node, runtime: MaybeMissing[Any], object_path: List[str]
+    stub: MaybeMissing[nodes.Node], runtime: MaybeMissing[Any], object_path: List[str]
 ) -> Iterator[Error]:
     """Entry point for comparing a stub to a runtime object.
 
@@ -206,28 +200,44 @@ def verify_mypyfile(
         yield Error(object_path, "is not a module", stub, runtime)
         return
 
-    # Check things in the stub that are public
+    # Check things in the stub
     to_check = set(
         m
         for m, o in stub.names.items()
-        if o.module_public and (not m.startswith("_") or hasattr(runtime, m))
+        if not o.module_hidden and (not m.startswith("_") or hasattr(runtime, m))
     )
-    runtime_public_contents = [
-        m
-        for m in dir(runtime)
-        if not m.startswith("_")
-        # Ensure that the object's module is `runtime`, e.g. so that we don't pick up reexported
-        # modules and infinitely recurse. Unfortunately, there's no way to detect an explicit
-        # reexport missing from the stubs (that isn't specified in __all__)
-        and getattr(getattr(runtime, m), "__module__", None) == runtime.__name__
-    ]
-    # Check all things declared in module's __all__, falling back to runtime_public_contents
-    to_check.update(getattr(runtime, "__all__", runtime_public_contents))
+
+    def _belongs_to_runtime(r: types.ModuleType, attr: str) -> bool:
+        obj = getattr(r, attr)
+        obj_mod = getattr(obj, "__module__", None)
+        if obj_mod is not None:
+            return obj_mod == r.__name__
+        return not isinstance(obj, types.ModuleType)
+
+    runtime_public_contents = (
+        runtime.__all__
+        if hasattr(runtime, "__all__")
+        else [
+            m
+            for m in dir(runtime)
+            if not m.startswith("_")
+            # Ensure that the object's module is `runtime`, since in the absence of __all__ we
+            # don't have a good way to detect re-exports at runtime.
+            and _belongs_to_runtime(runtime, m)
+        ]
+    )
+    # Check all things declared in module's __all__, falling back to our best guess
+    to_check.update(runtime_public_contents)
     to_check.difference_update({"__file__", "__doc__", "__name__", "__builtins__", "__package__"})
 
     for entry in sorted(to_check):
+        stub_entry = stub.names[entry].node if entry in stub.names else MISSING
+        if isinstance(stub_entry, nodes.MypyFile):
+            # Don't recursively check exported modules, since that leads to infinite recursion
+            continue
+        assert stub_entry is not None
         yield from verify(
-            stub.names[entry].node if entry in stub.names else MISSING,
+            stub_entry,
             getattr(runtime, entry, MISSING),
             object_path + [entry],
         )
@@ -257,8 +267,10 @@ def verify_typeinfo(
         mangled_entry = entry
         if entry.startswith("__") and not entry.endswith("__"):
             mangled_entry = "_{}{}".format(stub.name, entry)
+        stub_to_verify = next((t.names[entry].node for t in stub.mro if entry in t.names), MISSING)
+        assert stub_to_verify is not None
         yield from verify(
-            next((t.names[entry].node for t in stub.mro if entry in t.names), MISSING),
+            stub_to_verify,
             getattr(runtime, mangled_entry, MISSING),
             object_path + [entry],
         )
@@ -340,7 +352,7 @@ def _verify_arg_default_value(
 ) -> Iterator[str]:
     """Checks whether argument default values are compatible."""
     if runtime_arg.default != inspect.Parameter.empty:
-        if stub_arg.kind not in (nodes.ARG_OPT, nodes.ARG_NAMED_OPT):
+        if stub_arg.kind.is_required():
             yield (
                 'runtime argument "{}" has a default value but stub argument does not'.format(
                     runtime_arg.name
@@ -369,7 +381,7 @@ def _verify_arg_default_value(
                     )
                 )
     else:
-        if stub_arg.kind in (nodes.ARG_OPT, nodes.ARG_NAMED_OPT):
+        if stub_arg.kind.is_optional():
             yield (
                 'stub argument "{}" has a default value but runtime argument does not'.format(
                     stub_arg.variable.name
@@ -388,10 +400,10 @@ def maybe_strip_cls(name: str, args: List[nodes.Argument]) -> List[nodes.Argumen
 
 class Signature(Generic[T]):
     def __init__(self) -> None:
-        self.pos = []  # type: List[T]
-        self.kwonly = {}  # type: Dict[str, T]
-        self.varpos = None  # type: Optional[T]
-        self.varkw = None  # type: Optional[T]
+        self.pos: List[T] = []
+        self.kwonly: Dict[str, T] = {}
+        self.varpos: Optional[T] = None
+        self.varkw: Optional[T] = None
 
     def __str__(self) -> str:
         def get_name(arg: Any) -> str:
@@ -412,7 +424,7 @@ class Signature(Generic[T]):
             if isinstance(arg, inspect.Parameter):
                 return arg.default != inspect.Parameter.empty
             if isinstance(arg, nodes.Argument):
-                return arg.kind in (nodes.ARG_OPT, nodes.ARG_NAMED_OPT)
+                return arg.kind.is_optional()
             raise AssertionError
 
         def get_desc(arg: Any) -> str:
@@ -436,12 +448,12 @@ class Signature(Generic[T]):
 
     @staticmethod
     def from_funcitem(stub: nodes.FuncItem) -> "Signature[nodes.Argument]":
-        stub_sig = Signature()  # type: Signature[nodes.Argument]
+        stub_sig: Signature[nodes.Argument] = Signature()
         stub_args = maybe_strip_cls(stub.name, stub.arguments)
         for stub_arg in stub_args:
-            if stub_arg.kind in (nodes.ARG_POS, nodes.ARG_OPT):
+            if stub_arg.kind.is_positional():
                 stub_sig.pos.append(stub_arg)
-            elif stub_arg.kind in (nodes.ARG_NAMED, nodes.ARG_NAMED_OPT):
+            elif stub_arg.kind.is_named():
                 stub_sig.kwonly[stub_arg.variable.name] = stub_arg
             elif stub_arg.kind == nodes.ARG_STAR:
                 stub_sig.varpos = stub_arg
@@ -453,7 +465,7 @@ class Signature(Generic[T]):
 
     @staticmethod
     def from_inspect_signature(signature: inspect.Signature) -> "Signature[inspect.Parameter]":
-        runtime_sig = Signature()  # type: Signature[inspect.Parameter]
+        runtime_sig: Signature[inspect.Parameter] = Signature()
         for runtime_arg in signature.parameters.values():
             if runtime_arg.kind in (
                 inspect.Parameter.POSITIONAL_ONLY,
@@ -483,7 +495,7 @@ class Signature(Generic[T]):
         # For most dunder methods, just assume all args are positional-only
         assume_positional_only = is_dunder(stub.name, exclude_special=True)
 
-        all_args = {}  # type: Dict[str, List[Tuple[nodes.Argument, int]]]
+        all_args: Dict[str, List[Tuple[nodes.Argument, int]]] = {}
         for func in map(_resolve_funcitem_from_decorator, stub.items):
             assert func is not None
             args = maybe_strip_cls(stub.name, func.arguments)
@@ -508,7 +520,7 @@ class Signature(Generic[T]):
                 ]
                 return mypy.typeops.make_simplified_union([t for t in all_types if t])
 
-        def get_kind(arg_name: str) -> int:
+        def get_kind(arg_name: str) -> nodes.ArgKind:
             kinds = {arg.kind for arg, _ in all_args[arg_name]}
             if nodes.ARG_STAR in kinds:
                 return nodes.ARG_STAR
@@ -527,7 +539,7 @@ class Signature(Generic[T]):
                 return nodes.ARG_OPT if is_pos else nodes.ARG_NAMED_OPT
             return nodes.ARG_POS if is_pos else nodes.ARG_NAMED
 
-        sig = Signature()  # type: Signature[nodes.Argument]
+        sig: Signature[nodes.Argument] = Signature()
         for arg_name in sorted(all_args, key=get_position):
             # example_arg_name gives us a real name (in case we had a fake index-based name)
             example_arg_name = all_args[arg_name][0][0].variable.name
@@ -537,9 +549,9 @@ class Signature(Generic[T]):
                 initializer=None,
                 kind=get_kind(arg_name),
             )
-            if arg.kind in (nodes.ARG_POS, nodes.ARG_OPT):
+            if arg.kind.is_positional():
                 sig.pos.append(arg)
-            elif arg.kind in (nodes.ARG_NAMED, nodes.ARG_NAMED_OPT):
+            elif arg.kind.is_named():
                 sig.kwonly[arg.variable.name] = arg
             elif arg.kind == nodes.ARG_STAR:
                 sig.varpos = arg
@@ -652,11 +664,7 @@ def verify_funcitem(
         yield Error(object_path, "is not present at runtime", stub, runtime)
         return
 
-    if (
-        not isinstance(runtime, (types.FunctionType, types.BuiltinFunctionType))
-        and not isinstance(runtime, (types.MethodType, types.BuiltinMethodType))
-        and not inspect.ismethoddescriptor(runtime)
-    ):
+    if not is_probably_a_function(runtime):
         yield Error(object_path, "is not a function", stub, runtime)
         if not callable(runtime):
             return
@@ -664,11 +672,8 @@ def verify_funcitem(
     for message in _verify_static_class_methods(stub, runtime, object_path):
         yield Error(object_path, "is inconsistent, " + message, stub, runtime)
 
-    try:
-        signature = inspect.signature(runtime)
-    except (ValueError, RuntimeError):
-        # inspect.signature throws sometimes
-        # catch RuntimeError because of https://bugs.python.org/issue39504
+    signature = safe_inspect_signature(runtime)
+    if not signature:
         return
 
     stub_sig = Signature.from_funcitem(stub)
@@ -736,11 +741,7 @@ def verify_overloadedfuncdef(
         # We get here in cases of overloads from property.setter
         return
 
-    if (
-        not isinstance(runtime, (types.FunctionType, types.BuiltinFunctionType))
-        and not isinstance(runtime, (types.MethodType, types.BuiltinMethodType))
-        and not inspect.ismethoddescriptor(runtime)
-    ):
+    if not is_probably_a_function(runtime):
         yield Error(object_path, "is not a function", stub, runtime)
         if not callable(runtime):
             return
@@ -748,9 +749,8 @@ def verify_overloadedfuncdef(
     for message in _verify_static_class_methods(stub, runtime, object_path):
         yield Error(object_path, "is inconsistent, " + message, stub, runtime)
 
-    try:
-        signature = inspect.signature(runtime)
-    except ValueError:
+    signature = safe_inspect_signature(runtime)
+    if not signature:
         return
 
     stub_sig = Signature.from_overloadedfuncdef(stub)
@@ -839,7 +839,7 @@ def _resolve_funcitem_from_decorator(dec: nodes.OverloadPart) -> Optional[nodes.
         # anything else when running on typeshed's stdlib.
         return None
 
-    func = dec.func  # type: nodes.FuncItem
+    func: nodes.FuncItem = dec.func
     for decorator in dec.original_decorators:
         resulting_func = apply_decorator_to_funcitem(decorator, func)
         if resulting_func is None:
@@ -869,8 +869,31 @@ def verify_decorator(
 def verify_typealias(
     stub: nodes.TypeAlias, runtime: MaybeMissing[Any], object_path: List[str]
 ) -> Iterator[Error]:
-    if False:
-        yield None
+    if isinstance(runtime, Missing):
+        # ignore type aliases that don't have a runtime counterpart
+        return
+    stub_target = mypy.types.get_proper_type(stub.target)
+    if isinstance(stub_target, mypy.types.Instance):
+        yield from verify(stub_target.type, runtime, object_path)
+        return
+    if isinstance(stub_target, mypy.types.UnionType):
+        if not getattr(runtime, "__origin__", None) is Union:
+            yield Error(object_path, "is not a Union", stub, runtime, stub_desc=str(stub_target))
+        # could check Union contents here...
+        return
+    if isinstance(stub_target, mypy.types.TupleType):
+        if tuple not in getattr(runtime, "__mro__", ()):
+            yield Error(
+                object_path, "is not a subclass of tuple", stub, runtime,
+                stub_desc=str(stub_target)
+            )
+        # could check Tuple contents here...
+        return
+    if isinstance(stub_target, mypy.types.AnyType):
+        return
+    yield Error(
+        object_path, "is not a recognised type alias", stub, runtime, stub_desc=str(stub_target)
+    )
 
 
 SPECIAL_DUNDERS = ("__init__", "__new__", "__call__", "__init_subclass__", "__class_getitem__")
@@ -887,6 +910,25 @@ def is_dunder(name: str, exclude_special: bool = False) -> bool:
     return name.startswith("__") and name.endswith("__")
 
 
+def is_probably_a_function(runtime: Any) -> bool:
+    return (
+        isinstance(runtime, (types.FunctionType, types.BuiltinFunctionType))
+        or isinstance(runtime, (types.MethodType, types.BuiltinMethodType))
+        or (inspect.ismethoddescriptor(runtime) and callable(runtime))
+    )
+
+
+def safe_inspect_signature(runtime: Any) -> Optional[inspect.Signature]:
+    try:
+        return inspect.signature(runtime)
+    except Exception:
+        # inspect.signature throws ValueError all the time
+        # catch RuntimeError because of https://bugs.python.org/issue39504
+        # catch TypeError because of https://github.com/python/typeshed/pull/5762
+        # catch AttributeError because of inspect.signature(_curses.window.border)
+        return None
+
+
 def is_subtype_helper(left: mypy.types.Type, right: mypy.types.Type) -> bool:
     """Checks whether ``left`` is a subtype of ``right``."""
     left = mypy.types.get_proper_type(left)
@@ -929,8 +971,8 @@ def get_mypy_type_of_runtime_value(runtime: Any) -> Optional[mypy.types.Type]:
         type_info = builtins.names["function"].node
         assert isinstance(type_info, nodes.TypeInfo)
         fallback = mypy.types.Instance(type_info, [anytype()])
-        try:
-            signature = inspect.signature(runtime)
+        signature = safe_inspect_signature(runtime)
+        if signature:
             arg_types = []
             arg_kinds = []
             arg_names = []
@@ -952,7 +994,7 @@ def get_mypy_type_of_runtime_value(runtime: Any) -> Optional[mypy.types.Type]:
                     arg_kinds.append(nodes.ARG_STAR2)
                 else:
                     raise AssertionError
-        except ValueError:
+        else:
             arg_types = [anytype(), anytype()]
             arg_kinds = [nodes.ARG_STAR, nodes.ARG_STAR2]
             arg_names = [None, None]
@@ -999,7 +1041,7 @@ def get_mypy_type_of_runtime_value(runtime: Any) -> Optional[mypy.types.Type]:
         return fallback
 
 
-_all_stubs = {}  # type: Dict[str, nodes.MypyFile]
+_all_stubs: Dict[str, nodes.MypyFile] = {}
 
 
 def build_stubs(modules: List[str], options: Options, find_submodules: bool = False) -> List[str]:
@@ -1069,7 +1111,7 @@ def get_typeshed_stdlib_modules(custom_typeshed_dir: Optional[str]) -> List[str]
     """Returns a list of stdlib modules in typeshed (for current Python version)."""
     stdlib_py_versions = mypy.modulefinder.load_stdlib_py_versions(custom_typeshed_dir)
     packages = set()
-    # Typeshed doesn't cover Python 3.5.
+    # Typeshed's minimum supported Python 3 is Python 3.6
     if sys.version_info < (3, 6):
         version_info = (3, 6)
     else:
@@ -1198,7 +1240,11 @@ def parse_options(args: List[str]) -> argparse.Namespace:
         description="Compares stubs to objects introspected from the runtime."
     )
     parser.add_argument("modules", nargs="*", help="Modules to test")
-    parser.add_argument("--concise", action="store_true", help="Make output concise")
+    parser.add_argument(
+        "--concise",
+        action="store_true",
+        help="Makes stubtest's output more concise, one line per error",
+    )
     parser.add_argument(
         "--ignore-missing-stub",
         action="store_true",
@@ -1209,12 +1255,6 @@ def parse_options(args: List[str]) -> argparse.Namespace:
         action="store_true",
         help="Ignore errors for whether an argument should or shouldn't be positional-only",
     )
-    parser.add_argument(
-        "--custom-typeshed-dir", metavar="DIR", help="Use the custom typeshed in DIR"
-    )
-    parser.add_argument(
-        "--check-typeshed", action="store_true", help="Check all stdlib modules in typeshed"
-    )
     parser.add_argument(
         "--allowlist",
         "--whitelist",
@@ -1223,7 +1263,8 @@ def parse_options(args: List[str]) -> argparse.Namespace:
         default=[],
         help=(
             "Use file as an allowlist. Can be passed multiple times to combine multiple "
-            "allowlists. Allowlists can be created with --generate-allowlist"
+            "allowlists. Allowlists can be created with --generate-allowlist. Allowlists "
+            "support regular expressions."
         ),
     )
     parser.add_argument(
@@ -1238,19 +1279,20 @@ def parse_options(args: List[str]) -> argparse.Namespace:
         action="store_true",
         help="Ignore unused allowlist entries",
     )
-    config_group = parser.add_argument_group(
-        title='mypy config file',
-        description="Use a config file instead of command line arguments. "
-                    "Plugins and mypy path are the only supported "
-                    "configurations.",
-    )
-    config_group.add_argument(
-        '--mypy-config-file',
+    parser.add_argument(
+        "--mypy-config-file",
+        metavar="FILE",
         help=(
-            "An existing mypy configuration file, currently used by stubtest to help "
-            "determine mypy path and plugins"
+            "Use specified mypy config file to determine mypy plugins "
+            "and mypy path"
         ),
     )
+    parser.add_argument(
+        "--custom-typeshed-dir", metavar="DIR", help="Use the custom typeshed in DIR"
+    )
+    parser.add_argument(
+        "--check-typeshed", action="store_true", help="Check all stdlib modules in typeshed"
+    )
 
     return parser.parse_args(args)
 
